package org.netkernel.scheduler;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.netkernel.container.IKernel;
import org.netkernel.container.IKernelListener;
import org.netkernel.container.ILogger;
import org.netkernel.request.IAsyncRequestee;
import org.netkernel.request.IExpiry;
import org.netkernel.request.IRequest;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.IRequestor;
import org.netkernel.request.IResponse;
import org.netkernel.request.IResponseMeta;
import org.netkernel.request.impl.ExpiryFactory;
import org.netkernel.request.impl.MetaImpl;
import org.netkernel.request.impl.RequestFactory;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.request.impl.ResponseImpl;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.INetKernelException;
import org.netkernel.urii.IResolution;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.impl.NetKernelException;
import org.netkernel.util.AsynchResponseJoin;
import org.netkernel.util.ExclusionSet;
import org.netkernel.util.ExpiryIdentitySet;
import org.netkernel.util.MetaIdentitySet;
import org.netkernel.util.SynchronousRequestee;
import org.netkernel.util.Utils;

public class RequestState implements IRequestState, IRequestor
{
	private enum State {
		INITIAL,
		RESOLVE_PRIMARY,
		RESOLVE_PRIMARY_DEBUG,
		PRIMARY_ISSUE,
		PRIMARY_RESPONSE,
		PRIMARY_TEST,
		PRIMARY_BUSY,
		PRIMARY_RELEASED,
		PRIMARY_PENDING,
		TRANSREPT,
		RESOLVE_TRANSREPT,
		RESOLVE_TRANSREPT_DEBUG,
		TRANSREPT_ISSUE,
		TRANSREPT_BUSY,
		TRANSREPT_RELEASED,
		TRANSREPT_PENDING,
		RESPONSE_READY,
		CLEANUP,
		COMPLETE,
		RESOLVE_DIRECT,
		RESOLVE_DIRECT_DEBUG,
		RESOLVE_DIRECT_COMPLETE,
		ADDITIONAL_DEPENDENCY
	}

	private static final IRequestState[] INITIAL_SUBSTATES=new IRequestState[0];
	private static final IExpiry[] EMPTY_EXPIRY_ARRAY = new IExpiry[0];
	private static final IExpiry[] EXPIRED_EXPIRY_ARRAY = new IExpiry[] { ExpiryFactory.getAlwaysExpired() };

	private IRequestState mParent;
	protected RootState mRoot;
	protected IRequest mRequestIn;
	
	private State mState;
	private IRequestState[] mSubStates=INITIAL_SUBSTATES;
	private int mSubStateIndex=0;
	private int mTransreptSubStateIndex=0;
	private Thread mExecutionThread;
	private long mInitTime;
	private long mStartTime;
	private int mElapsedTime;
	private int mTotalTime;
	private AsynchResponseJoin mResponseJoin;
	private boolean mHandlingErrorAlready;
	private IResponse mResponse;
	private IResponseMeta mResponseMeta;
	private IResponse mPrimaryResponse;
	private int mScope = Integer.MAX_VALUE;
	private IRequest mPrimaryRequest;
	private boolean mNeedsCache;
	private boolean mFromCache;
	private boolean mIgnoreDependencies=false;
	private IExpiry[] mExpiryDeterminants=EMPTY_EXPIRY_ARRAY;
	
	private IAsyncRequestee mRequesteeOut;
	private IRequest mRequestOut;
	
	/** request to be resolved **/
	private IRequest mRequestToResolve;
	/** set of excluded spaces, used to optimise resolve */
	protected ExclusionSet mResolveExcludedSpaces;
	/** pointer to current scope during resolve */
	protected IRequestScopeLevel mResolveScope;
	/** list of durable spaces to be placed into request scope during resolve */
	protected List<ISpace> mResolveDurableSpaces;
	/** accumulated debug info */
	protected TraceDebugCapture mResolveDebug;
	/** the endpoint resolved in last resolve */
	protected IEndpoint mResolvedEndpoint;
	/** the request resolved in the last resolve */
	protected IRequest mResolvedRequest;
	/** the logical endpoint id resolved in the last resolve */
	protected IIdentifier mResolvedEndpointId;
	
	public RequestState(IRequestState aParent, IRequest aRequestIn)
	{	mParent=aParent;
		mRoot=aParent.getRoot();
		mRequestIn=aRequestIn;
		mState=State.INITIAL;
	}
	
	public RootState getRoot()
	{	return mRoot;
	}
	
	private IRequest getRootRequest()
	{	return mRoot.getRootRequest();
	}
	
	public IRequest getRequestIn()
	{	return mRequestIn;
	}
	
	protected Scheduler getScheduler()
	{	return mRoot.getScheduler();
	}

	protected IKernel getKernel()
	{	return mRoot.getScheduler().getKernel();
	}
	
	public void declareSubrequestIssue(IRequestState aSubState)
	{	//doesn't need to be threadsafe because always called by thread initiating subrequest
		if (mIgnoreDependencies || mSubStates==null)
		{	//caused by parent being complete or not tracking dependencies
			//ignore
		}
		else
		{
			if (mSubStates==INITIAL_SUBSTATES)
			{	mSubStates=new IRequestState[8];
			}
			else if (mSubStateIndex==mSubStates.length)
			{	IRequestState[] newSubStates =new IRequestState[mSubStateIndex*2];
				System.arraycopy(mSubStates, 0, newSubStates, 0, mSubStateIndex);
				mSubStates=newSubStates;
			}
			mSubStates[mSubStateIndex++]=aSubState;
		}
	}

	public void declareSubrequestResponse(IRequestState aSubstate)
	{
	}
	
	
	public long getTimestamp()
	{	return mInitTime;
	}
	
	public Thread getExecutionThread()
	{	return mExecutionThread;
	}
	
	public IRequestState getParent()
	{	return mParent;
	}
	
	public State getState()
	{	return mState;
	}
	
	public List<RequestState> getSubStates()
	{	List<RequestState> result;
		IRequestState[] substates=mSubStates;
		if (substates==null)
		{	result = Collections.EMPTY_LIST;
		}
		else
		{	int len=mSubStateIndex;
			if (len>substates.length) len=substates.length;
			result = new ArrayList(len);
			int i=0;
			while (i<len && substates[i]!=null)
			{	if (substates[i] instanceof RequestState)
				{	result.add((RequestState)substates[i++]);
				}
			}
		}
		return result;
	}
	
	/** Return true if the state is stuck without hope of returning a response
	 */
	public boolean isResponseless()
	{	
		boolean result=mResponseJoin==null || !mResponseJoin.isAccepting();
		
		if (result)
		{	switch (mState)
			{	case PRIMARY_PENDING:
				case TRANSREPT_PENDING:
				case RESOLVE_PRIMARY:
				case RESOLVE_PRIMARY_DEBUG:
				case RESOLVE_TRANSREPT:
				case RESOLVE_TRANSREPT_DEBUG:
				case RESOLVE_DIRECT:
				case RESOLVE_DIRECT_DEBUG:
					break;
				default:
					result=false;
			}
		}

		IRequestState[] substates=mSubStates;
		if (result && substates!=null)
		{
			int i=0;
			while (i<substates.length && substates[i]!=null && result)
			{	RequestState rs=(RequestState)substates[i];
				if (rs.getState()!=State.COMPLETE)
				{	result=false;
				}
				i++;
			}
		}
		return result;
	}
	
	public void pauseTimer()
	{	if (mStartTime!=0) //0 is timer hasn't been start because it's a resolve request
		{	long d=getRelativeTime()-mStartTime;
			mStartTime=0;
			mElapsedTime+=(int)d;
		}
	}
	
	public void resumeTimer()
	{	mStartTime=getRelativeTime();
	}	
	
	/** Return total time and reset clock
	 */
	public int getElapsedTime()
	{	int result=mElapsedTime;
		mTotalTime+=mElapsedTime;
		mElapsedTime=0;
		return result;
	}
	
	public int getTotalTime()
	{	return mTotalTime+mElapsedTime;
	}
	
	private long getRelativeTime()
	{	return getScheduler().getRelativeTime();
	}
	
	private void setExceptionResponse(Throwable aThrowable)
	{	setExceptionResponse(aThrowable, MetaImpl.EXPIRED);
	}
	
	public boolean isComplete()
	{	return mState==State.COMPLETE || mState==State.CLEANUP;
	}
	
	public IInternalResponseMeta getResponseMeta()
	{	return (IInternalResponseMeta)mResponseMeta;
	}
	
	private void declareScope(int aScope)
	{	if (aScope<mScope)
		{	mScope=aScope;
		}
	}
	private int getUnresolvedScope()
	{	return mScope;
	}
	
	void setExceptionResponse(Throwable aThrowable, IResponseMeta aMeta)
	{	Throwable th;
		if (aThrowable instanceof java.lang.Error)
		{	th = new RequestFrameError(mRequestIn,aThrowable);
		}
		else
		{	th = new RequestFrameException(mRequestIn,aThrowable);
		}
		mResponse = new ResponseImpl(mRequestIn, th, aMeta);
		if (mState==State.PRIMARY_PENDING)
		{	mPrimaryResponse=mResponse;
			mTransreptSubStateIndex=mSubStateIndex;
		}
		mState=State.RESPONSE_READY;
	}
	
	protected void initialiseResolve(IRequest aRequestToResolve)
	{	mRequestToResolve=aRequestToResolve;
		mResolveScope = aRequestToResolve.getRequestScope();
		mResolveExcludedSpaces = new ExclusionSet(16);
		mResolveDurableSpaces = new ArrayList<ISpace>(16);
		mResolveDebug=null;
		mResponse=null;
	}	

	/******************************************************************
	 * Set a request that we need to issue next
	 * @param aRequestOut
	 * @param aRequesteeOut
	 ******************************************************************/
	private void declareRequestOut(IRequest aRequestOut, IAsyncRequestee aRequesteeOut)
	{	mRequestOut=aRequestOut;
		mRequesteeOut=aRequesteeOut;
	}
	
	public String toString()
	{	return "requestState("+mState.name()+", "+RequestFactory.toShortString(mRequestIn)+")";
	}
	
	/******************************************************************
	 * processRequest - method to do work towards processing the
	 * request
	 ******************************************************************/
	
	//private List<State> mStateHistory = new ArrayList<State>();
	public void processRequest()
	{	mExecutionThread=Thread.currentThread();
		try
		{	do 
			{	//mStateHistory.add(mState);
				switch (mState)
				{	case INITIAL:
						stateInitial();
						break;
					case RESPONSE_READY:
						stateResponseReady();
						break;
					case RESOLVE_PRIMARY:
						stateResolve(false, State.PRIMARY_TEST, State.RESOLVE_PRIMARY_DEBUG);
						break;						
					case RESOLVE_PRIMARY_DEBUG:
						stateResolve(true, State.RESPONSE_READY, State.RESPONSE_READY);
						break;						
					case PRIMARY_TEST:
						statePrimaryTest();
						break;						
					case PRIMARY_ISSUE:
						statePrimaryIssue();
						break;						
					case PRIMARY_RELEASED:
						statePrimaryReleased();
						break;
					case PRIMARY_RESPONSE:
						statePrimaryResponse();
						break;
					case TRANSREPT:
						stateTransrept();
						break;
					case RESOLVE_TRANSREPT:
						stateResolve(false, State.TRANSREPT_ISSUE, State.RESOLVE_TRANSREPT_DEBUG);
						break;						
					case RESOLVE_TRANSREPT_DEBUG:
						stateResolve(true, State.RESPONSE_READY, State.RESPONSE_READY);
						break;						
					case TRANSREPT_ISSUE:
						stateTransreptIssue();
						break;
					case TRANSREPT_RELEASED:
						stateTransreptReleased();
						break;
					case RESOLVE_DIRECT:
						stateResolveDirect();
						break;
					case RESOLVE_DIRECT_COMPLETE:
						stateResolveComplete();
						break;
					case CLEANUP:
						stateCleanup();
						break;
				}
			
				if (mRequestOut!=null)
				{	IRequest request=mRequestOut;
					mRequestOut=null;
					IKernel k=getRoot().getScheduler().getKernel();
					
					//ensure space is commissioned
					if (!request.getRequestScope().getSpace().isCommissioned())
					{	INetKernelException ex=Utils.createFormattedException("EX_UNCOMMISSIONED_SPACE","MSG_UNCOMMISSIONED_SPACE",k.getLogger(),request.getRequestScope().getSpace());
						setExceptionResponse(ex);
					}
					else
					{	if (request.getVerb()!=IRequest.VERB_RESOLVE)
						{	k.getKernelListener().endpointInvoke(getRootRequest(),mParent.getRequestIn(),mRequestIn,request,(IEndpoint)mRequesteeOut);
							resumeTimer();
						}	
						
						mResponseJoin=new AsynchResponseJoin();
						Thread.currentThread().setContextClassLoader(request.getRequestScope().getSpace().getClassLoader());
						
						mRequesteeOut.onAsyncRequest(request);
						IResponse synchResponse = mResponseJoin.stopCapture();					
						
						if (synchResponse!=null)
						{	handleResponse(synchResponse);
						}
						else
						{	break;
						}
					}
				}
			} while (mState!=State.COMPLETE && mState!=State.PRIMARY_BUSY && mState!=State.TRANSREPT_BUSY);
		}
		catch (Throwable th)
		{	IKernel k=getRoot().getScheduler().getKernel();

			if (th instanceof OutOfMemoryError)
			{	k.getRepresentationCache().clear();
				System.gc();
				OutOfMemoryError oome=(OutOfMemoryError)th;
				getScheduler().getKernel().getLogger().log(ILogger.SEVERE, this, "EX_OUT_OF_MEMORY",th.getMessage());
			}
			if (mHandlingErrorAlready)
			{	//th.printStackTrace();
				throw new IllegalStateException();
			}
			else
			{
				try
				{	mHandlingErrorAlready=true;
					unhandledException(th);
				}
				catch (IllegalStateException th2)
				{	// error handling unexpected error - log cause
					StringWriter sw=new StringWriter(512);
					PrintWriter pw = new PrintWriter(sw,true);
					try
					{	th.printStackTrace(pw);
					} catch (Throwable t) {;}
					pw.flush();
					String trace=sw.toString();
					k.getLogger().log(ILogger.SEVERE, this, "EX_UNHANDLED_KERNEL_ERROR", trace);
					setExceptionResponse(th);
					mState=State.CLEANUP;
					mRequestIn.getRequestor().onResponse(mResponse);
				}
			}
		}
		mExecutionThread=null;
	}
	
	/******************************************************************
	 * Deal with an unhandled exception whilst processing
	 * @param aUnhandled the exception we need to handle
	 ******************************************************************/
	private void unhandledException(Throwable aUnhandled)
	{	
		//need to free any blocked endpoint
		if (mRequesteeOut!=null)
		{	if (mRequesteeOut instanceof IEndpoint && !((IEndpoint)mRequesteeOut).isThreadSafe())
			{	releaseEndpoint((IEndpoint)mRequesteeOut,this);
			}
		}
		
		if (mState==State.PRIMARY_PENDING || mState==State.TRANSREPT_PENDING)
		{ //unhandled exception in endpoint
			if (aUnhandled instanceof java.lang.Exception)
			{	SubrequestException sre=new SubrequestException(mResolvedRequest,mResolvedEndpoint,mResolvedEndpointId, aUnhandled);
				setExceptionResponse(sre);
				processRequest();
			}
			else if (aUnhandled instanceof java.lang.Error)
			{	SubrequestError sre=new SubrequestError(mResolvedRequest,mResolvedEndpoint,mResolvedEndpointId, aUnhandled);
				//aUnhandled.printStackTrace();
				setExceptionResponse(sre); // unhandled never cached
				Thread.interrupted(); // called to clear after a thread kill
				processRequest();
			}			
		}
		else if (mState!=State.COMPLETE)
		{	setExceptionResponse(aUnhandled);
			processRequest();
		}
		else
		{	// error in returning result to parent
			Scheduler s = getScheduler();
			RequestState parentState = s.getParentState(mRequestIn);
			if (parentState!=null)
			{	parentState.setExceptionResponse(aUnhandled);
				s.getRequestTable().put(parentState);
			}
		}
	}	
	
	/******************************************************************
	 * Release endpoint and queue next request to execute if necessary
	 * @param aEndpoint 
	 * @param aRequestState 
	 ******************************************************************/
	private void releaseEndpoint(IEndpoint aEndpoint, RequestState aRequestState)
	{	Scheduler s=getScheduler();
		RequestState released = s.getLockedEndpoints().unlock(aEndpoint,aRequestState);
		if (released!=null)
		{	if (released.mState==State.PRIMARY_BUSY)
			{	released.mState=State.PRIMARY_RELEASED;
				s.getRequestTable().put(released);
			}
			else if (released.mState==State.TRANSREPT_BUSY)
			{	released.mState=State.TRANSREPT_RELEASED;
				s.getRequestTable().put(released);
			}
			else
			{	s.getKernel().getLogger().log(ILogger.SEVERE,this,"ERR_SCHEDULER_UNSTABLE_BLOCKED");
			}
		}
	}
	
	/******************************************************************
	 * onResponse - called when an asynchronous response to issued request is received.
	 * It will either store response for future processing or handle inline
	 * @param aResponse
	 ******************************************************************/
	public void onResponse(IResponse aResponse)
	{	
		if (aResponse.getRequest().getVerb()!=IRequest.VERB_RESOLVE)
		{	pauseTimer();
			IEndpoint endpoint = (IEndpoint)mRequesteeOut;
			// free up a request blocked endpoint that's just responded
		    if (!endpoint.isThreadSafe())
		    {	releaseEndpoint(endpoint,this);
		    }
		}
	
		// handle response now if we've missed doing it synchronously
		// we'll only miss response if we don't return the response during the initial onRequest()
		// - i.e. for accessors that return response async
		Integer handleStatus=mResponseJoin.acceptResponse(aResponse);
		if (handleStatus==AsynchResponseJoin.HANDLE_ASYNC)
		{	handleResponse(aResponse);
			processRequest();
		}
		else if (handleStatus==AsynchResponseJoin.HANDLE_ASYNC_DOUBLE)
		{	IRequest requestOut = aResponse.getRequest();		
			getScheduler().getKernel().getLogger().log(ILogger.SEVERE, this, "MSG_UNEXPECTED_RESPONSE", RequestFactory.toShortString(requestOut), mState.name());
		}
	}	
	
	/******************************************************************
	 * Actually handling of response from sub-request regardless of when
	 * @param aResponse
	 * @return true if we should continue to process response
	 ******************************************************************/
	private void handleResponse(IResponse aResponse)
	{
		IRequest requestOut = aResponse.getRequest();		
		Object rep = aResponse.getRepresentation();
				
		if (rep instanceof java.lang.Throwable )
		{	if (rep instanceof java.lang.Exception)
			{	SubrequestException sre=new SubrequestException(requestOut,mRequesteeOut,mResolvedEndpointId, (Exception)rep);
				setExceptionResponse(sre,aResponse.getMeta());
			}
			else
			{	SubrequestError sre=new SubrequestError(requestOut,mRequesteeOut,mResolvedEndpointId, (Error)rep);
				setExceptionResponse( sre,aResponse.getMeta());
			}
		}
		else
		{	// a non-exception result
			switch(mState)
			{	case PRIMARY_PENDING:
					mResponse=aResponse;
					mPrimaryResponse=mResponse;
					mState=State.PRIMARY_RESPONSE;
					break;
				case TRANSREPT_PENDING:
					mResponse=aResponse;
					mState=State.RESPONSE_READY;
					break;
				case RESOLVE_PRIMARY:
				case RESOLVE_PRIMARY_DEBUG:
				case RESOLVE_TRANSREPT:
				case RESOLVE_TRANSREPT_DEBUG:
				case RESOLVE_DIRECT:
				case RESOLVE_DIRECT_DEBUG:
					// in these cases we just re-enter the same state and use the response to determine what to do next
					mResponse=aResponse;
					break;
				default:
					// should never happen
					break;
			}			
		}
	}		
	
	/******************************************************************
	 * kill - cleanup state regardless of current state
	 * 
	 ******************************************************************/
	public void kill(String aKillExceptionId)
	{
		getKernel().getLogger().log(ILogger.WARNING, this, "MSG_KILL_REQUEST", this);
		
		// free up any request blocked endpoint
		if (mRequesteeOut instanceof IEndpoint)
		{	IEndpoint endpoint = (IEndpoint)mRequesteeOut;
		    if (!endpoint.isThreadSafe())
		    {	releaseEndpoint(endpoint,this);
		    }
		}
		// issue dummy exception response
		if (mState!=State.RESPONSE_READY && mState!=State.COMPLETE)
		{	setExceptionResponse(new NetKernelException(aKillExceptionId));
			processRequest();
		}
	}
	
	
	/******************************************************************
	 * stateInitial - state code for  determining initial processing of incoming request
	 * 
	 ******************************************************************/
	private void stateInitial()
	{	
		mInitTime=getScheduler().getAbsoluteTime();
	
		IRequest request=mRequestIn;
		mIgnoreDependencies=request.getHeaders().getValue(IRequest.HEADER_FORGET_DEPENDENCIES)!=null;
		
		int verb=request.getVerb();
		if (verb==IRequest.VERB_TRANSREPT)
		{	Object responseIn = mRequestIn.getValue(IRequest.TYPE_SOURCED_RESPONSE);
			if (responseIn==null || !(responseIn instanceof IResponse))
			{	mResponse=Utils.generateExceptionResponse(mRequestIn,getKernel(),"EX_BAD_TRANSREPT");
				mState=State.RESPONSE_READY;
			}
			else
			{	mResponse=(IResponse)responseIn;
				mState=State.TRANSREPT;
			}
		}
		
		else if (verb==IRequest.VERB_RESOLVE)
		{	IRequest toResolve = (IRequest)mRequestIn.getValue(IRequest.TYPE_RESOLVE_REQUEST);
			initialiseResolve(toResolve);
			mState=State.RESOLVE_DIRECT;
		}
		
		else // SOURCE SINK EXISTS DELETE NEW META
		{	
			IResponse response=null;
			// test cache for existence of response to unresolved request
			verb=mRequestIn.getVerb();
			boolean cache=mRequestIn.getHeaders().getValue(IRequest.HEADER_NO_CACHE)==null;
			if (cache && (verb==IRequest.VERB_SOURCE || verb==IRequest.VERB_META || verb==IRequest.VERB_EXISTS))
			{	response=getKernel().getRepresentationCache().get(request);
				if (response!=null)
				{	mResponse=response;
					mFromCache=true;
					getKernel().getKernelListener().cacheHit(getRootRequest(),mParent.getRequestIn(),mRequestIn,mResponse);
					mState=State.RESPONSE_READY;	
				}
			}
			
			if (response==null)
			{	Object[] cachedResolutionResult = getKernel().getResolutionCache().getResolution(request);
				if (cachedResolutionResult!=null)
				{	IResolution cachedResolution = (IResolution)cachedResolutionResult[0];
					mResolvedEndpoint=cachedResolution.getEndpoint();
					mResolvedEndpointId=cachedResolution.getEndpointIdentifier();
					mResolvedRequest=RequestFactory.cloneWithScopeAndRequestor(request, cachedResolution.getScope(), this, cachedResolution.getEndpointIdentifier());
					int unresolvedScope = ((Integer)cachedResolutionResult[1]).intValue();
					declareScope(unresolvedScope); //scope from resolution cache
					mState=State.PRIMARY_ISSUE;
				}
				else
				{	// initiate resolve of primary endpoint
					initialiseResolve(mRequestIn);
					mState=State.RESOLVE_PRIMARY;
				}
			}
		}

	}
	
	/******************************************************************
	 * stateResolvePrimary - state code for  
	 ******************************************************************/
	private void stateResolve(boolean aDebug, State aSuccessState, State aFailureState)
	{	
		IRequest toBeResolved;
		if (mResponse!=null)
		{	// response back from last resolve, test it 
			Object rep=mResponse.getRepresentation();
			if (rep!=null)
			{	
				if (! (rep instanceof IResolution))
				{	Exception ex = Utils.createFormattedException("EX_BAD_RESOLVE_RESPONSE", "MSG_BAD_RESOLVE_RESPONSE", getKernel().getLogger(), mResolveScope);
					mResponse=Utils.generateExceptionResponse(mRequestToResolve, ex);
					mState=State.RESPONSE_READY;
					return;
				}
			}
			
			IResolution lastResolution=(IResolution)rep;
			if (lastResolution!=null && lastResolution.getEndpoint()!=null)
			{	// resolved!
				int unresolvedScope = mResolveScope.getDepth();
				declareScope(unresolvedScope);
				mResolvedEndpoint=lastResolution.getEndpoint();
				mResolvedEndpointId=lastResolution.getEndpointIdentifier();
				mResolvedRequest=RequestFactory.cloneWithScopeAndRequestor(mRequestToResolve,lastResolution.getScope(),this,mResolvedEndpointId);
				if (!mResponse.getMeta().getExpiry().isExpired(getScheduler().getApproximateTime()))
				{	getKernel().getResolutionCache().putResolution(mRequestToResolve,getUnresolvedScope(),lastResolution);
				}
				mState=aSuccessState;
				mResolveDurableSpaces=null;
				mResolveExcludedSpaces=null;
				return;
			}
			else
			{	// not resolved yet, descope
				if (mResolveScope.isDurable())
				{	mResolveDurableSpaces.add(mResolveScope.getSpace());
				}
				IRequestScopeLevel lastResolveScope=mResolveScope;
				mResolveScope=mResolveScope.getParent();
				ILogger logger=getKernel().getLogger();				
				
				boolean earlyTerminate=lastResolution!=null;
				if (mResolveScope!=null && !earlyTerminate)
				{	
					IRequestScopeLevel resolveScope=mResolveScope;
					for (int i=mResolveDurableSpaces.size()-1; i>=0; i--)
					{	resolveScope=RequestScopeLevelImpl.createDurableScopeLevel((ISpace)mResolveDurableSpaces.get(i),resolveScope);
					}
					
					if (mResolveScope!=resolveScope)
					{	resolveScope = RequestScopeLevelImpl.appendScopeLevel(mResolveScope,resolveScope);
					}
					toBeResolved = RequestFactory.cloneWithScope(mRequestToResolve,resolveScope);
							
					if (aDebug)
					{	mResolveDebug.add(logger.format("RESOLVE_PARENT", mResolveScope.getSpace()));
					}
				}
				else
				{	// no resolution possible
					if (aDebug)
					{	//declareUnresolvedScope(-1);
						//TODO indicate that resolution failure through lack of scope should resolve
						if (earlyTerminate)
						{	mResolveDebug.add(logger.format("RESOLVE_STOP"));
							declareScope(lastResolveScope.getDepth());
						}
						else
						{	mResolveDebug.add(logger.format("RESOLVE_NONE"));
							// it depends upon the non-existence of a scope level above root
							declareScope(lastResolveScope.getDepth()-1);
						}
						//mResolveDebug.add(logger.format(earlyTerminate?"RESOLVE_STOP":"RESOLVE_NONE"));
						ResolutionFailureException rfe = new ResolutionFailureException(getKernel(),mRequestToResolve,mResolveDebug);
						setExceptionResponse(rfe,MetaImpl.DEPENDENT_EXPIRY);
						IKernelListener kl = getKernel().getKernelListener();
						kl.resolutionFailure(getRootRequest(),mParent.getRequestIn(),mRequestIn,mResponse);
					}
					else
					{	// do resolve again in debug mode 
						mResolveExcludedSpaces.clear();
						mResolveScope = mRequestToResolve.getRequestScope();
						mResolveDurableSpaces.clear();
						mResolveDebug = new TraceDebugCapture();
						mResponse=null;
				
						mResolveDebug.add(logger.format("RESOLVE_STARTING", mResolveScope.getSpace()));								
					}
					mState=aFailureState;
					return;
				}				
			}
		}
		else
		{	// initial innermost resolve 
			toBeResolved=mRequestToResolve;
		}
		
		
		IRequest request=RequestFactory.createResolveRequest(toBeResolved,mResolveExcludedSpaces,mResolveDebug,mRequestIn,this);
		declareRequestOut(request,mResolveScope.getSpace());
	}
	
	/******************************************************************
	 * statePrimaryTest - called after a primary resolve. Tests cache and if no resolved response
	 * is available it will issue request.
	 ******************************************************************/
	private void statePrimaryTest()
	{	//mPrimaryEndpoint=mResolvedEndpoint;
		//mPrimaryRequest=mResolvedRequest;
		
		// see if resolved request is found in cache
		boolean cache=mRequestIn.getHeaders().getValue(IRequest.HEADER_NO_CACHE)==null;
		if (cache)
		{	IResponse response=getKernel().getRepresentationCache().get(mResolvedRequest);
			if (response!=null)
			{	int scope=((IInternalResponseMeta)response.getMeta()).getResolvedScope();
				scope=(scope<getUnresolvedScope())?scope:getUnresolvedScope();
				getKernel().getRepresentationCache().putAdditionalUnresolvedScope(mRequestIn.getRequestScope(),scope,mResolvedRequest);
				IInternalResponseMeta meta = (IInternalResponseMeta)response.getMeta();
				if (meta.getUnresolvedScope()!=scope)
				{	meta = IInternalResponseMeta.cloneWithScope(meta,scope);
				}
				mResponse=new ResponseImpl(mRequestIn,response.getRepresentation(),meta);
				
				mFromCache=true;
				getKernel().getKernelListener().cacheHit(getRootRequest(),mParent.getRequestIn(),mRequestIn,mResponse);
				mState=State.RESPONSE_READY;
			}
			else
			{	statePrimaryIssue();
			}
		}
		else
		{	statePrimaryIssue();
		}
	}

	/******************************************************************
	 * statePrimaryIssue - state code for when primary endpoint is ready to be executed we
	 * must decide if it can accept the request right now
	 ******************************************************************/
	private void statePrimaryIssue()
	{	if (mResolvedEndpoint.isThreadSafe() || mResolvedRequest.getVerb()==IRequest.VERB_RESOLVE || mResolvedRequest.getVerb()==IRequest.VERB_META || getScheduler().getLockedEndpoints().lock(mResolvedEndpoint,this) )
		{	mState=State.PRIMARY_PENDING;
			declareRequestOut(mResolvedRequest,mResolvedEndpoint);
		}
		else
		{	mState=State.PRIMARY_BUSY;
		}
		mPrimaryRequest=mResolvedRequest;
	}
	
	/******************************************************************
	 * statePrimaryReleased - state code for  when primary endpoint is released after
	 * being busy
	 ******************************************************************/
	private void statePrimaryReleased()
	{	mState=State.PRIMARY_PENDING;
		declareRequestOut(mResolvedRequest,mResolvedEndpoint);
	}
	
	/******************************************************************
	 * statePrimaryResponse - state code for when response is returned from primary. We
	 * test it to see if it needs transreption
	 ******************************************************************/
	@SuppressWarnings("unchecked")
	private void statePrimaryResponse()
	{	mNeedsCache=true;
		mTransreptSubStateIndex=mSubStateIndex;
		Object rep = mResponse.getRepresentation();
		if (rep==null || mRequestIn.getRepresentationClass().isAssignableFrom(rep.getClass()))
		{	// representation is suitable without transreption
			mState=State.RESPONSE_READY;
		}
		else
		{	// representation requires transreption
			long duration = getElapsedTime();
			getKernel().getKernelListener().endpointResponse(getRootRequest(),mRequestIn, (IEndpoint)mRequesteeOut,duration,mResponse);
			mState=State.TRANSREPT;
		}
	}
	
	/******************************************************************
	 * stateTransrept - state code for  
	 * 
	 ******************************************************************/
	private void stateTransrept()
	{	IRequest req=RequestFactory.createTransreptRequest(mRequestIn,mResponse,this);
		initialiseResolve(req);
		mState=State.RESOLVE_TRANSREPT;	
	}
	
	/******************************************************************
	 * stateTransreptComplete - state code for  
	 * 
	 ******************************************************************/
	private void stateTransreptIssue()
	{	if (mResolvedEndpoint.isThreadSafe() || getScheduler().getLockedEndpoints().lock(mResolvedEndpoint,this) )
		{	mState=State.TRANSREPT_PENDING;
			declareRequestOut(mResolvedRequest,mResolvedEndpoint);
		}
		else
		{	mState=State.TRANSREPT_BUSY;
		}		
	}
	
	/******************************************************************
	 * stateTransreptReleased - state code for  
	 * 
	 ******************************************************************/
	private void stateTransreptReleased()
	{	mState=State.TRANSREPT_PENDING;
		declareRequestOut(mResolvedRequest,mResolvedEndpoint);
	}
	
	/******************************************************************
	 * stateResolveDirect - state code for  
	 * 
	 ******************************************************************/
	private void stateResolveDirect()
	{	
		boolean debug = mRequestIn.getValue(IRequest.TYPE_RESOLVE_DEBUG)!=null;
		if (debug && mResolveDebug==null)
		{	mResolveDebug = new TraceDebugCapture();
			IRequest toResolve = (IRequest)mRequestIn.getValue(IRequest.TYPE_RESOLVE_REQUEST);
			mResolveDebug.add(getKernel().getLogger().format("RESOLVE_STARTING", toResolve.getRequestScope().getSpace()));								
		}
		stateResolve(debug,State.RESOLVE_DIRECT_COMPLETE,State.RESOLVE_DIRECT_COMPLETE);
	}

	/******************************************************************
	 * stateResolveComplete - state code for completed direct resolve
	 * 
	 ******************************************************************/
	private void stateResolveComplete()
	{	TraceDebugCapture debug=(TraceDebugCapture)mRequestIn.getValue(IRequest.TYPE_RESOLVE_DEBUG);
		if (debug!=null)
		{	// add resolve debug to request if wanted
			debug.cloneFrom(mResolveDebug);
		}
		if (mResponse==null || mResponse.getRepresentation() instanceof ResolutionFailureException)
		{	mResponse=new ResponseImpl(mRequestIn,null,MetaImpl.EXPIRED);
		}
		mState=State.RESPONSE_READY;
	}
	
	/******************************************************************
	 * stateResponseReady - state code for return a response back to requestor 
	 * 
	 ******************************************************************/
	 void stateResponseReady()
	{	if (!mFromCache)
		{	IInternalResponseMeta meta;
			IRequestState[] states = mSubStates;
			if (states!=null)
			{	meta=generateResponseMeta(states);
			}
			else
			{	meta=new InternalResponseMeta(ExpiryFactory.getAlwaysExpired(),mResponse.getMeta().getUserMetaData(),0,0,0,0,IInternalResponseMeta.NONE, 0 ,EMPTY_EXPIRY_ARRAY);
			}
			mResponse = new ResponseImpl(mRequestIn,mResponse.getRepresentation(), meta );
			if (mRequestIn.getVerb()!=IRequest.VERB_RESOLVE && mRequesteeOut instanceof IEndpoint)
			{	getKernel().getKernelListener().endpointResponse(getRootRequest(),mRequestIn, (IEndpoint)mRequesteeOut,getElapsedTime(),mResponse);
			}			
		}
		if (mNeedsCache)
		{	boolean cache=mRequestIn.getHeaders().getValue(IRequest.HEADER_NO_CACHE)==null;
			if (cache)
			{	getKernel().getRepresentationCache().put(mResponse,mPrimaryRequest);
			}
		}
		
		mParent.declareSubrequestResponse(this);
		
		IRequestor requestor = mRequestIn.getRequestor();
		mResponseMeta=mResponse.getMeta();
		mState=State.CLEANUP;

		if (!(requestor instanceof SynchronousRequestee))
		{	//asynchronous request
			mParent.resumeTimer();
			IRequest parent=mRequestIn.getParent();
			if (parent!=null)
			{	ClassLoader cl=parent.getRequestScope().getSpace().getClassLoader();
				if (cl!=null)
				{	Thread.currentThread().setContextClassLoader(cl);
				}
			}
		}
		
		requestor.onResponse(mResponse);
		mResponse=null;
	}
	 
	/******************************************************************
	 * stateResponseReady - state code for return a response back to requestor 
	 * 
	 ******************************************************************/
	private void stateCleanup()
	{	mSubStates=null;
		mExpiryDeterminants=null;
		mPrimaryRequest=null;
		mPrimaryResponse=null;
		mRequestToResolve=null;
		mResolveDebug=null;
		mResolvedEndpoint=null;
		mResolvedRequest=null;
		mResolveDurableSpaces=null;
		mResolveExcludedSpaces=null;
		mResolveScope=null;
		mResponse=null;
		
		//keep these for sake of kernel view
		//mRequestIn=null;
		//mRoot=null;
		
		//keep this so endpoint can be unlocked if needed after late async response
		//mRequesteeOut=null;

		//keep this in case we get any async response
		//mResponseJoin=null;
		//mParent=null;
		
		mState=State.COMPLETE;
	}
	 
	/******************************************************************
	 * generateResponseMeta - process response with all dependencies for cost and
	 *  expiry accounted for
	 ******************************************************************/
	 private IInternalResponseMeta generateResponseMeta(IRequestState[] aStates)
	 {
		 //snapshot substates including additional dependencies now 
		 int ss=mSubStateIndex;
		 int maxExpiries=1;
		 int maxCosts=0;
		 int transreptSubStateIndex=0;
		 List<IRequestState> substates = new ArrayList<IRequestState>(ss);
		 if (mPrimaryResponse!=null)
		 {	IRequestResponseFields usermeta=mPrimaryResponse.getMeta().getUserMetaData();
		 	List<IResponse> additionalDependencies = (List<IResponse>)usermeta.getValue(IResponseMeta.KEY_ADDITIONAL_DEPENDENCIES);
			 if (additionalDependencies!=null)
			{	for (int i=0; i<additionalDependencies.size(); i++)
				{	IResponse response = additionalDependencies.get(i);
					IRequestState state=new AdditionalDependencyState(response);
					substates.add(state);
					maxExpiries+=state.getResponseMeta().getExpiryDeterminants().length;
					maxCosts+=1+state.getResponseMeta().getCostDeterminants().length;
				}
				additionalDependencies.clear();
			}
		 }
		 transreptSubStateIndex=substates.size();
		 for (int i=0; i<ss; i++)
		 {	IRequestState state=aStates[i];
			 if (state.isComplete())
		 	{	substates.add(state);
		 		IInternalResponseMeta meta=state.getResponseMeta();
		 		maxExpiries+=meta.getExpiryDeterminants().length;
		 		maxCosts+=1+meta.getCostDeterminantsLength();
		 	}
	 		if (i+1==mTransreptSubStateIndex)
	 		{	transreptSubStateIndex=substates.size();
	 		}
		 }
		 if (mResponse!=mPrimaryResponse)
		 {	IRequestResponseFields usermeta=mResponse.getMeta().getUserMetaData();
		 	List<IResponse> additionalDependencies = (List<IResponse>)usermeta.getValue(IResponseMeta.KEY_ADDITIONAL_DEPENDENCIES);
			if (additionalDependencies!=null)
			{	for (int i=0; i<additionalDependencies.size(); i++)
				{	IResponse response = additionalDependencies.get(i);
					IRequestState state=new AdditionalDependencyState(response);
					substates.add(state);
					maxExpiries+=state.getResponseMeta().getExpiryDeterminants().length;
					maxCosts+=1+state.getResponseMeta().getCostDeterminantsLength();
				}
			}
			
		 }
		 
		//generate cost structure and resolved scope
		int localCost=getTotalTime();
		int costThreshold=(int)getKernel().getTimer().getRelativeTimeDivisor();
		MetaIdentitySet ourCostDeterminants = new MetaIdentitySet(maxCosts);
		int resolvedScope=Integer.MAX_VALUE;
		for (IRequestState state : substates)
		{	IInternalResponseMeta meta=state.getResponseMeta();
			int cost=meta.getLocalCost();
			if (cost>costThreshold)
			{	ourCostDeterminants.add(meta);
			}
			else
			{	localCost+=cost;
			}
			IInternalResponseMeta[] cds=meta.getCostDeterminants();
			int cdl=meta.getCostDeterminantsLength();
			for (int i=0; i<cdl; i++)
			{	ourCostDeterminants.add(cds[i]);
			}
			/*
			for (IInternalResponseMeta sm: meta.getCostDeterminants())
			{	if (sm==null) break;
				ourCostDeterminants.add(sm);
			}
			*/
			int scope=meta.getUnresolvedScope();
			if (scope<resolvedScope)
			{	resolvedScope=scope;
			}
		}
		int unresolvedScope = resolvedScope<mScope?resolvedScope:mScope;
		
		IExpiry expiryOut;
		// determine expiry for primary
		long now=getScheduler().getApproximateTime();
		if (mIgnoreDependencies)
		{	expiryOut=ExpiryFactory.getAlwaysExpired();
			mExpiryDeterminants = EXPIRED_EXPIRY_ARRAY;
		}
		else
		{		
			if (mPrimaryResponse!=null)
			{	IExpiry primaryExpiry=mPrimaryResponse.getMeta().getExpiry();
				if (ExpiryFactory.tracksDependencies(primaryExpiry))
				{	IExpiry[] primaryDeterminants;
					if (transreptSubStateIndex>0)
					{	primaryDeterminants=mergeSubStateExpiry(0, transreptSubStateIndex, maxExpiries, substates).getAsArray();
					}
					else
					{	primaryDeterminants=EMPTY_EXPIRY_ARRAY;
					}
					// this next method optimally sets mExpiryDeterminants as a side-effect
					expiryOut=rationalizeDependencyExpiry(primaryExpiry, primaryDeterminants);					
				}
				else
				{	expiryOut=primaryExpiry;
					mExpiryDeterminants = new IExpiry[] {expiryOut};
				}
			}
			else
			{	expiryOut=ExpiryFactory.getNeverExpired();
				mExpiryDeterminants = new IExpiry[0];
			}
			
			// determine expiry for transrept
			if (mResponse!=mPrimaryResponse)
			{	IExpiry transreptExpiry=mResponse.getMeta().getExpiry();
				if (ExpiryFactory.tracksDependencies(transreptExpiry))
				{	IExpiry[] transreptDeterminants;
					if (mTransreptSubStateIndex<mSubStateIndex)
					{	ExpiryIdentitySet transreptDeterminantsSet=mergeSubStateExpiry(transreptSubStateIndex, mSubStateIndex, maxExpiries, substates);
						//add expiry determinants from primary
						for (int i=0; i<mExpiryDeterminants.length; i++)
						{	transreptDeterminantsSet.add(mExpiryDeterminants[i]);
						}
						transreptDeterminants=transreptDeterminantsSet.getAsArray();
					}
					else
					{	transreptDeterminants=mExpiryDeterminants;
					}
					// this next method optimally sets mExpiryDeterminants as a side-effect
					expiryOut=rationalizeDependencyExpiry(transreptExpiry, transreptDeterminants);
				}
				else
				{	expiryOut=transreptExpiry;
					mExpiryDeterminants = new IExpiry[] {expiryOut};
				}				
			}
			
			if (ExpiryFactory.getType(expiryOut)!=ExpiryFactory.EXPIRY_ALWAYS && expiryOut.isExpired(now))
			{	//replace an abitrary expiry structure with always expired if it is expired, because once expired it can't change its mind
				expiryOut=ExpiryFactory.getAlwaysExpired();
				mExpiryDeterminants = EXPIRED_EXPIRY_ARRAY;
			}
		}
		
		int totalCost=localCost+ourCostDeterminants.getTotalCost();
		IInternalResponseMeta result=new InternalResponseMeta(expiryOut,mResponse.getMeta().getUserMetaData(),unresolvedScope,resolvedScope,localCost,totalCost,ourCostDeterminants.getAsArray(),ourCostDeterminants.size(),mExpiryDeterminants);
		return result;
	 }

	 private IExpiry rationalizeDependencyExpiry(IExpiry aExpiryIn, IExpiry[] aDeterminants)
	{	IExpiry result;
		IExpiry[] determinants=null;
	
		int typeIn=ExpiryFactory.getType(aExpiryIn);
		if (aDeterminants.length==0 && typeIn==ExpiryFactory.EXPIRY_DEPENDENT)
		{	// this common corner case can lead to efficiencies
			result = ExpiryFactory.getNeverExpired();
			determinants=EMPTY_EXPIRY_ARRAY;
		}
		else if (aDeterminants.length==1 && typeIn==ExpiryFactory.EXPIRY_DEPENDENT && ExpiryFactory.getType(aDeterminants[0])==ExpiryFactory.EXPIRY_ALWAYS)
		{	// this common corner case can lead to efficiencies
			result = ExpiryFactory.getAlwaysExpired();
			determinants=EXPIRED_EXPIRY_ARRAY;
		}
		else
		{	switch(typeIn)
			{	case ExpiryFactory.EXPIRY_DEPENDENT:
					result = ExpiryFactory.getDependentExpiry(aDeterminants);
					determinants=aDeterminants;
					break;
				 case ExpiryFactory.EXPIRY_MIN_CONSTANT_DEPENDENT:
					 result = ExpiryFactory.getMinConstantDependentExpiry(ExpiryFactory.getTime(aExpiryIn),aDeterminants);
					 break;
				 case ExpiryFactory.EXPIRY_MAX_CONSTANT_DEPENDENT:
					 result = ExpiryFactory.getMaxConstantDependentExpiry(ExpiryFactory.getTime(aExpiryIn),aDeterminants);
					 break;
				 case ExpiryFactory.EXPIRY_MIN_FUNCTION_DEPENDENT:
					 if (aDeterminants.length>0)
					 {	result = ExpiryFactory.getMinFunctionDependentExpiry(ExpiryFactory.getFunction(aExpiryIn),aDeterminants);
					 }
					 else
					 {	result = ExpiryFactory.getFunction(aExpiryIn);
					 }
					 break;
				default:
					result=ExpiryFactory.getAlwaysExpired();
					break;
			}
			if (determinants==null)
			{	determinants=new IExpiry[]{result};
			}
		}
		mExpiryDeterminants=determinants;
		return result;
	}	
	
	private static ExpiryIdentitySet mergeSubStateExpiry(int i0, int i1, int maxExpiries, List<IRequestState> aStates)
	{
		ExpiryIdentitySet	determinantsSet= new ExpiryIdentitySet(maxExpiries);
		for (int i=i0; i<i1; i++)
		{	IRequestState state=aStates.get(i);
			IExpiry[] expiryDeterminants = state.getResponseMeta().getExpiryDeterminants();
			for (int j=expiryDeterminants.length-1; j>=0; j--)
			{	IExpiry ed=expiryDeterminants[j];
				int type=ExpiryFactory.getType(ed);
				if (type!=ExpiryFactory.EXPIRY_NEVER && type!=ExpiryFactory.EXPIRY_DEPENDENT)
				{	if (type==ExpiryFactory.EXPIRY_ALWAYS)
					{	determinantsSet.clear();
						determinantsSet.add(ed);
						break;
					}
					else
					{	determinantsSet.add(ed);
					}
				}
			}
		}
		return determinantsSet;
	}
	
}
